#define FALSE (0)
#define TRUE (!FALSE)

/*
** Converts TAIP TM times and dates from message value to correct value.
** Works on all TAIP messages from all receivers through the end of year 2017.
** If values are already correct, they are passed back unchanged.
** Returns FALSE if date has been processed successfully.
** Returns TRUE if input date is not a possible TAIP TM date.
*/
short ConvertPostWNRODate (
   short *year,
   short *month,
   short *day,
   short *hour,
   short *min,
	short *sec)
{
	short
   	ctr,
      leapday;
	const	short
   	cumdim[15] =
			{0, -59, -28, 0, 31, 61, 92, 122, 153, 184, 214, 245, 275, 306, 337},
      divday[5] = {0,0,20,21,6},
      posdaybias[5] = {0,0,11,9,10},
      negdaybias[5] = {0,0,20,21,0};

/*
** Possible TM dates are all dates from 6 jan 1980 through the year 2017 and
** 55 days in the months of feb, mar, and apr of the year 2116
*/
	if ((*year) < 1999)
   {

      leapday = (*month == 2) && (*day == 29);

	   /*
      ** compute apparent days since 1 mar 1976
      */
  		(*day) = (*day) + 365*((*year) - 1976) + cumdim[(*month)];

      if ((*day) < 1407)
      {
      	return TRUE;
      }

		/*
      ** then add add one (leap) day for every four years = 1461 days
         ** max four cycles; could be replaced by divide
      */
      ctr = (*day) - 1460;
      while (ctr > 0)
      {
       	(*day)++;
         ctr -= 1460;
      }
      if (leapday)
      {
         (*day)--;
      }
      /*
      ** then add 1024 weeks for second GPS epoch and at the same time
	   ** subtract 5*1461 to base time at 01 mar 1996 instead of 01 mar 1980
      ** -137 = (1024*7 - 5*1461)
      */
  		(*day) -= 137;
   	(*year) = 1996;

   	/*
      ** (*day) is now days since 01 mar 1996
   	** (*month) is months since march
	   ** (*year) = year
      ** max four cycles; could be replaced by divide
      */
		while ((*day) >= 1461)
  		{
			(*day) -= 1461;
			(*year) += 4;
	  	}

		if((*day) == 0)
		{
			/*
         ** leap day - add another four years to (*year) & return
         */
			(*day) = 29;
			(*month) = 2;
		}
   	else
	   {
   		/*
         ** max three cycles; could be replaced by divide
         */
			while ((*day) > 365)
			{
				(*day) -= 365;
				(*year) ++;
			}
			(*month) = 14;

     		/*
         ** max twelve cycles; could be replaced by efficient sort
         */
			while ((*day) <= cumdim[(*month)])
      	{
	      	(*month)--;
   	   }
	     	(*day) -= cumdim[(*month)];

			if ((*month) > 12)
   	   {
				(*year)++;
				(*month) -= 12;
			}

		}
      /* no change to hour, minute, second */
      return FALSE;
   }
   else if ((*year) == 2116)
   /*
   ** 8 weeks of strange TM messages between 22 aug and 15 oct 1999
   ** output dates will always be between 11 feb and 6 apr 2116
   ** both time and date are biased
   */
   {
#define TAIPSECBIAS 16
#define TAIPMINBIAS 28
#define TAIPHOURBIAS 6

	   *year = 1999;
      if (*month > 4)
      {
      	return TRUE;
      }
      /*
      ** compute correct hour, minute, second;
      ** no leap second will occur during this short time period
      */
      (*sec) -= TAIPSECBIAS;
      if ((*sec) < 0)
      {
      	(*sec)+= 60;
         (*min)--;
      }
      (*min) -= TAIPMINBIAS;
      if ((*min) < 0)
      {
      	(*min)+=60;
         (*hour)--;
      }
      (*hour) -= TAIPHOURBIAS;
      if ((*hour) < 0)
      {
      	(*hour)+=24;
         (*day)--;
      }

      /* note - value of zero allowed for (*day) in the next equaitons */

      /* map into correct month & day */
     	if ((*day) <= divday[(*month)])
      {
      	(*day) += posdaybias[(*month)];
	      (*month) += 6;
      }
      else
      {
      	(*day) -= negdaybias[(*month)];
	      (*month) += 7;
      }
		return FALSE;
   }
   else if ((*year) < 2016)
   {
   	/*
      ** TAIP message OK
      */
		return FALSE;
   }
	return TRUE;
}


/***********************************************************************/
/***********************************************************************/
/**********        driver for unit test      ***************************/
/***********************************************************************/
/***********************************************************************/

const short mdays[] = { 31, 30, 31, 30, 31, 31, 30, 31, 30, 31, 31, 28 };
#define DAY 86400L

void gps2zulu(
	unsigned long utc_secs,
   short *yr, short *mo, short *day,
   short *hh, short *mm, short *ss )
{
	short epoch;

    /*
	*	Work day out of 4-year epoch, starts Mar 1, ends on Feb 29.  The
	*	first epoch is Mar 1 1980 (JD 61).  GPS day 0 is Jan 6, 1980 (JD 6)
	*/

    utc_secs -= (55 * DAY);

	epoch = (short) ((utc_secs / DAY) / (365*4 + 1));
    *day = (short) ((utc_secs / DAY) % (365*4 + 1));

	*yr = 1980 + epoch * 4;

	if (*day == 365*4) {	/* leap year */
		*yr += 3;
		*mo = 11;		/* at this point, both m & d are indexes */
		*day = 28;		/* which will be incremented later to offset from 1 */
	}
	else {
		*yr += *day/365;
		*day %= 365;
		for (*mo = 0; *mo < 12; (*mo)++) {
			if (*day < mdays[*mo]) {
				break;
			}
			else {
				*day -= mdays[*mo];
			}
		}
	}

	/*
	*	Adjust mo/yr to remove epoch offset
	*/

	if (*mo >= 10) {		/* index 10 is Jan */
		(*yr)++;
		*mo -= 9;
	}
	else {
		*mo += 3;
	}

	(*day)++;	/* start days @ 1 */

    *hh = (short) ((utc_secs /3600L) % 24L);
    *mm = (short) ((utc_secs /60L) % 60L);
	*ss = (short) (utc_secs % 60L);
}

#include <stdlib.h>
#include <stdio.h>
/**/
FILE
	*fdebug;

void main (void)
{

	short
   	day, month, year,
   	inday, inmonth, inyear,
   	sec, min, hr,
		gday, gmonth, gyear,
   	gsec, gmin, ghr,
      oldgmonth;
   unsigned long
   	utctime0,
   	utctime,
      nowtime;
   long
   	numerrs,
   	deltai,
      numsecs,
   	secs;
	long
   	timeofday;
   char
   	dstr[200];


   fdebug = fopen ("fdebug.txt", "w");
   if (fdebug == NULL)
   {
   	printf ("\n could not open fdebug.txt");
   	exit (0);
   }

   utctime0 = 1024L*604800L - 9*86400L;
   deltai = 59L;
   numsecs = 366L*19*86400L;

	gps2zulu(utctime0, &gyear, &gmonth, &gday, &ghr, &gmin, &gsec);
   fprintf (fdebug, "\nTAIP Time & Date Test");
   fprintf (fdebug, "\nStart at %02d:%02d:%02d %02d/%02d/%4d ",
     	ghr, gmin, gsec, gday, gmonth, gyear);


  	numerrs=0;
   oldgmonth = gmonth;
   for (secs=0; secs<numsecs; secs+=deltai)
   {
   	utctime = utctime0 + secs;

		gps2zulu(utctime, &gyear, &gmonth, &gday, &ghr, &gmin, &gsec);
   	if (gyear > 2017) break;

      if (oldgmonth != gmonth)
      {
      	printf ("\n%02d/%02d/%04d", gday, gmonth, gyear);
      }
      oldgmonth = gmonth;

  		timeofday = (long)gsec + 60*(long)gmin + 3600L*(long)ghr;

   	nowtime = utctime % (1024*604800L); /* convert to GPS week */

		gps2zulu(nowtime, &year, &month, &day, &hr, &min, &sec);

  		timeofday = (long)sec + 60*(long)min + 3600L*(long)hr;

      sprintf (dstr, " %02d:%02d:%02d (=%6ld) %02d/%02d/%4d ",
      	hr, min, sec, timeofday, day, month, year);

  		ConvertPostWNRODate (&year, &month, &day, &hr, &min, &sec);

		if ((ghr-hr)||(gmin-min)||(gsec-sec)||(gday-day)||(gmonth-month)||(gyear-year))
      {
      	numerrs++;
	      fprintf (fdebug, "\npt %ld: true %02d:%02d:%02d %02d/%02d/%4d ",
   	   	secs, ghr, gmin, gsec, gday, gmonth, gyear);
         fprintf (fdebug, "%s", dstr);
         fprintf (fdebug, "calc %02d:%02d:%02d (=%6ld) %02d/%02d/%4d ",
   	   	hr, min, sec, timeofday, day, month, year);
  	      fprintf (fdebug, "  (%02d,%02d,%02d,%02d,%02d,%4d)",
   	   	ghr-hr, gmin-min, gsec-sec, gday-day, gmonth-month, gyear-year);
      }
   }
   fprintf (fdebug, "\nNumber of points = %ld, number of errors = %ld",
   	secs/deltai, numerrs);
	gps2zulu(utctime, &gyear, &gmonth, &gday, &ghr, &gmin, &gsec);
   fprintf (fdebug, "\nEnd at %02d:%02d:%02d %02d/%02d/%4d ",
     	ghr, gmin, gsec, gday, gmonth, gyear);

   fprintf (fdebug, "\nExtra NMEA Date Test from 6 jan 1980 through 29 feb 1980");
   gday = 22;
   gmonth = 8;
   gyear = 1999;
   inday = 6;
   inmonth = 1;
   inyear = 1980;
   ghr = 12;
   gmin = 0;
   gsec = 0;
   numerrs = 0;
   for (inday = 6; inmonth < 3; inday++, gday++)
   {
      if (((inmonth==1) && (inday>31)) || ((inmonth==2) && (inday>29)))
      {
      	/* rollover into the next month */
      	inday = 1;
         inmonth++;
      }

      if (((gmonth==8) && (gday>31)) || ((gmonth==9) && (gday>30)))
      {
      	/* rollover into the next month */
      	gday = 1;
         gmonth++;
      }

      year = inyear;
      month = inmonth;
      day = inday;
      hr = ghr;
      min = gmin;
      sec = gsec;

      sprintf (dstr, "inp %02d:%02d:%02d %02d/%02d/%4d ",
      	hr, min, sec, inday, inmonth, inyear);

  		ConvertPostWNRODate (&year, &month, &day, &hr, &min, &sec);

		if ((ghr-hr)||(gmin-min)||(gsec-sec)||(gday-day)||(gmonth-month)||(gyear-year))
      {
	      fprintf (fdebug, "\ntrue %02d:%02d:%02d %02d/%02d/%4d ",
  		   	ghr, gmin, gsec, gday, gmonth, gyear);
      	fprintf (fdebug, "%s", dstr);
	      fprintf (fdebug, "calc %02d:%02d:%02d %02d/%02d/%4d ",
   	   	hr, min, sec, day, month, year);
  	      fprintf (fdebug, "  (%02d,%02d,%02d,%02d,%02d,%4d)",
   	   	ghr-hr, gmin-min, gsec-sec, gday-day, gmonth-month, gyear-year);
      }
   }
 	fprintf (fdebug, "NMEA date errors: %ld", numerrs);
}



